package HesiodLDAP;

use warnings;
use strict;
use Carp;
use Net::LDAP;
use Socket qw(inet_aton AF_INET );

use version; our $VERSION = qv('1.0.0');

# exports
use base qw( Exporter );
our @EXPORT    = qw( resolve );
our @EXPORT_OK = qw(
    get_accounts_hashref
    get_groups_hashref
    get_hosts_hashref
    get_mounts_hashref
);
our %EXPORT_TAGS = (
    ALL => [ @EXPORT, @EXPORT_OK ],    # if :ALL requested
);

# constants
my $LDAPSERV  = 'ldap.ncsu.edu';
my $LDAPBASE  = 'dc=ncsu,dc=edu';
my $ISNUMERIC = qr{ \A \d+ \z }xms;

# external globals
our $last_error;
our @types =                           # this order is important
    qw(
    passwd
    filsys
    pobox
    gid
    uid
    grplist
    sloc
    cluster
    group
    pcap
    service
    lpralias
);

# internal variables
my %is_valid_type = map { $_ => 1 } @types;
my %resolve_sub   = ();
my $ldap_handle   = undef;
my $ldap_cache    = {};

########################################
# Usage         : my $hes = resolve( $identity, $type );
# Purpose       : Resolve the requested type data for the given identity
# Returns       : A string result. For one type, just return the data.
#               : For all types, prefix the data with the type.
# Parameters    : $identity = a userid, uid_number, or a hostname
#               : $type = which type to lookup. May be omitted.

sub resolve {
    my $identity = shift;
    my $type     = shift;

    # always start with a clear error var
    $last_error = undef;

    # we must have an identity
    if ( defined $identity ) {
        $identity = lc($identity);

        if ( $identity =~ /[^\w\-.]/ ) {
            $last_error = "identity contains invalid characters";
            return;
        }
    }
    else {
        $last_error = "identity was not provided";
        return;
    }

    # if we have a type, it must be known
    if ( defined $type ) {
        $type = lc($type);
        if ( !$is_valid_type{$type} ) {
            $last_error = "type '$type' is not recognized";
            return;
        }
        return &{ $resolve_sub{$type} }($identity);
    }

    # type not defined, try all of them in order
    my $result = '';
    foreach my $type (@types) {
        my $val = &{ $resolve_sub{$type} }($identity);
        if ($val) {
            foreach my $line ( split /\n/, $val ) {
                $result .= sprintf "%10s: %s\n", uc($type), $line;
            }
        }
        else {

            # quietly skip failures
        }
    }
    return $result;
}

########################################
# a resolve_sub needs to be defined for every type.
# so create a placeholder sub for all of them,
# then replace for each specific case
#
# this does not set last_error, so it will not overwrite more
# significant messages when used as one of a list of types

my $undef_sub = sub { return; };
foreach my $type (@types) {
    $resolve_sub{$type} = $undef_sub;
}

########################################
# Usage         : my $result = &{$resolve_sub{'cluster'}}($identity);
# Purpose       : Resolve the cluster entry for the given identity.
# Returns       : A data string.
# Parameters    : $identity = a hostname

$resolve_sub{'cluster'} = sub {
    my $identity = shift;

    # cluster is not defined for numeric ids
    return if ( $identity =~ $ISNUMERIC );

    # treat identity as a (partial) hostname
    my $hostname    = undef;
    my $packed_addr = inet_aton($identity);
    if ($packed_addr) {
        $hostname = gethostbyaddr( $packed_addr, AF_INET );
    }
    if ( !$hostname ) {
        $last_error = "failed to resolve $identity as a hostname";
        return;
    }

    # now we can assume we have a fully-qualified hostname
    my $href = get_hosts_hashref($hostname);

    # ldap didn't return a result, keep the $last_error and just quit
    return if ( !$href );

    # this result will have up to two parts
    my @res = ();
    if ( $href->{ncsuAllowNetgroup} ) {
        my $aref = $href->{ncsuAllowNetgroup};
        if ( !ref $aref ) {
            $aref = [$aref];
        }
        my $output = 'acl';
        foreach my $dn (@$aref) {
            my $stripped = _strip_dn($dn);
            if ($stripped) {
                $output .= ' ' . $stripped;
            }
        }
        push @res, $output;
    }
    if ( $href->{ncsuDenyNetgroup} ) {
        my $aref = $href->{ncsuDenyNetgroup};
        if ( !ref $aref ) {
            $aref = [$aref];
        }
        my $output = 'nacl';
        foreach my $dn (@$aref) {
            my $stripped = _strip_dn($dn);
            if ($stripped) {
                $output .= ' ' . $stripped;
            }
        }
        push @res, $output;
    }

    if (@res) {
        return join( "\n", @res );
    }
    else {
        $last_error
            = "ldap data missing: hosts: ncsuaAllowNetgroup, ncsuDenyNetgroup";
        return;
    }
};

########################################
# Usage         : my $result = &{$resolve_sub{'filsys'}}($identity);
# Purpose       : Resolve the filsys entry for the given identity.
# Returns       : A data string.
# Parameters    : $identity = a uid, uidNumber, or a hostname

$resolve_sub{'filsys'} = sub {
    my $identity = shift;

    # filsys is not defined for numeric ids
    return if ( $identity =~ $ISNUMERIC );

    # assume we have a userid or uid number
    my $href = get_accounts_hashref($identity);
    if ( $href && $href->{uid} && $href->{homeDirectory} ) {
        my $result
            = 'AFS ' . $href->{homeDirectory} . ' w /ncsu/' . $href->{uid};
        return $result;
    }

    # ldap returned, but values were missing
    if ($href) {
        $last_error = "ldap data missing: accounts: uid, homeDirectory";
        return;
    }

    # no result, maybe its a mount name
    $href = get_mounts_hashref($identity);
    if ( $href && $href->{ncsuAFSPath} && $href->{ncsuMountPath} ) {
        my $result
            = 'AFS ' . $href->{ncsuAFSPath} . ' w ' . $href->{ncsuMountPath};
        return $result;
    }

    # ldap returned, but values were missing
    if ($href) {
        $last_error = "ldap data missing: mounts: ncsuAFSPath, ncsuMountPath";
        return;
    }

    # ldap didn't return a result, keep the $last_error and just quit
    return;
};

########################################
# Usage         : my $result = &{$resolve_sub{'group'}}($identity, $numeric);
# Purpose       : Resolve the group entry for the given identity.
# Returns       : A data string.
# Parameters    : $identity = a gid or gidNumber
#               : $numeric = reverse the numeric test if true

$resolve_sub{'group'} = sub {
    my $identity = shift;
    my $numeric  = shift;

    # code is the same, but the logic is to only allow two cases
    # (numeric,gid) and (nonnumeric,group)
    if ( $identity =~ $ISNUMERIC ) {
        return if ( !$numeric );
    }
    else {
        return if ($numeric);
    }

    # assume we have a groupid or gid number
    my $href = get_groups_hashref($identity);
    if ( $href && $href->{cn} && $href->{gidNumber} ) {
        my $result = $href->{cn} . ':*:' . $href->{gidNumber} . ':';
        return $result;
    }

    # ldap returned, but values were missing
    if ($href) {
        $last_error = "ldap data missing: groups: cn, gidNumber";
        return;
    }

    # ldap didn't return a result, keep the $last_error and just quit
    return;
};

# 'gid' type is an alias for 'group', with the numeric test reversed
$resolve_sub{'gid'} = sub {
    return &{ $resolve_sub{'group'} }( $_[0], 1 );
};

########################################
# Usage         : my $result = &{$resolve_sub{'grplist'}}($identity);
# Purpose       : Resolve the grplist entry for the given identity.
# Returns       : A data string.
# Parameters    : $identity = a uid or uidNumber

$resolve_sub{'grplist'} = sub {
    my $identity = shift;

    # grplist is not defined for numeric ids
    return if ( $identity =~ $ISNUMERIC );

    # we should have an account uid
    my $href = get_accounts_hashref($identity);
    if ( $href && $href->{memberNisNetgroup} ) {
        my @groups = ();
        my $aref   = $href->{memberNisNetgroup};
        if ( !ref $aref ) {
            $aref = [$aref];
        }
        foreach my $ldap_grp (@$aref) {
            my $grp_name = _strip_dn($ldap_grp);
            next if ( !$grp_name );
            my $grp_href = get_groups_hashref($grp_name);
            next if ( !$grp_href || !$grp_href->{gidNumber} );
            push @groups, [ $grp_name, $grp_href->{gidNumber} ];
        }
        my $result = join ':', map {@$_}
            sort { $a->[1] <=> $b->[1] } @groups;
        return $result;
    }

    # ldap returned, but values were missing
    if ($href) {
        $last_error = "ldap data missing: accounts: memberNisNetgroup";
        return;
    }

    # ldap didn't return a result, keep the $last_error and just quit
    return;
};

########################################
# Usage         : my $result = &{$resolve_sub{'passwd'}}($identity, $numeric);
# Purpose       : Resolve the passwd entry for the given identity.
# Returns       : A data string.
# Parameters    : $identity = a uid, or uidNumber
#               : $numeric = reverse the numeric test if true

$resolve_sub{'passwd'} = sub {
    my $identity = shift;
    my $numeric  = shift;

    # code is the same, but the logic is to only allow two cases
    # (numeric,uid) and (nonnumeric,passwd)
    if ( $identity =~ $ISNUMERIC ) {
        return if ( !$numeric );
    }
    else {
        return if ($numeric);
    }

    # we should have a uid or uidNumber only
    my $href = get_accounts_hashref($identity);

    # ldap didn't return a result, keep the $last_error and just quit
    return if ( !$href );

    # assemble the result from whatever we have (missing results ignored)
    my @fields = ();
    foreach my $field (qw( uid uidNumber gidNumber gecos uid loginShell)) {
        my $val = $href->{$field} || '';
        if ( ref $val ) {
            $val = join ' ', @$val;
        }
        push @fields, $val;
    }

    # sneak in a * for the passwd
    splice( @fields, 1, 0, '*' );

    # fix the home path
    $fields[5] = '/ncsu/' . $fields[5];

    # result is the list of fields catenated with colons
    return join ':', @fields;
};

# 'uid' type is an alias for 'passwd', with the numeric test reversed
$resolve_sub{'uid'} = sub {
    return &{ $resolve_sub{'passwd'} }( $_[0], 1 );
};

########################################
# Usage         : my $result = &{$resolve_sub{'pobox'}}($identity);
# Purpose       : Resolve the pobox entry for the given identity.
# Returns       : A data string.
# Parameters    : $identity = a uid or uidNumber

$resolve_sub{'pobox'} = sub {
    my $identity = shift;

    # pobox is not defined for numeric ids
    return if ( $identity =~ $ISNUMERIC );

    # we should have a uid or uidNumber only
    my $href = get_accounts_hashref($identity);
    if ( $href && $href->{uid} && $href->{mailHost} ) {
        my $result = 'POP ' . $href->{mailHost} . ' ' . $href->{uid};
        return $result;
    }

    # ldap returned, but values were missing
    if ($href) {
        $last_error = "ldap data missing: accounts: uid, mailHost";
        return;
    }

    # ldap didn't return a result, keep the $last_error and just quit
    return;
};

########################################
# Usage         : my $href = get_accounts_hashref($identity);
# Purpose       : Perform an LDAP accounts query for the given identity
# Returns       : A hash reference of the search results.
# Parameters    : $identity = a uid or uidNumber

sub get_accounts_hashref {
    my $identity = shift;

    # switch the query on uid or uidNumber
    my $ou = 'ou=accounts';
    my $query;
    if ( $identity =~ $ISNUMERIC ) {
        $query = "(uidNumber=$identity)";
    }
    else {
        $query = "(uid=$identity)";
    }

    # do the query
    return _get_ldap_as_hashref( $ou, $query );
}

########################################
# Usage         : my $href = get_hosts_hashref($identity);
# Purpose       : Perform an LDAP hosts query for the given identity
# Returns       : A hash reference of the search results.
# Parameters    : $identity = a hostname

sub get_hosts_hashref {
    my $identity = shift;

    my $ou    = 'ou=hosts';
    my $query = '(&(objectClass=ncsuHost)' . "(cn=$identity)" . ')';

    # do the query
    return _get_ldap_as_hashref( $ou, $query );
}

########################################
# Usage         : my $href = get_groups_hashref($identity);
# Purpose       : Perform an LDAP groups query for the given identity
# Returns       : A hash reference of the search results.
# Parameters    : $identity = a gid or gidNumber

sub get_groups_hashref {
    my $identity = shift;

    # switch the query on gid or gidNumber
    my $ou = 'ou=access,ou=groups';
    my $query;
    if ( $identity =~ $ISNUMERIC ) {
        $query = "(gidNumber=$identity)";
    }
    else {
        $query = "(cn=$identity)";
    }

    # do the query
    return _get_ldap_as_hashref( $ou, $query );
}

########################################
# Usage         : my $href = get_mounts_hashref($identity);
# Purpose       : Perform an LDAP mounts query for the given identity
# Returns       : A hash reference of the search results.
# Parameters    : $identity = a mount name

sub get_mounts_hashref {
    my $identity = shift;

    my $ou    = 'ou=mounts';
    my $query = "(cn=$identity)";

    # do the query
    return _get_ldap_as_hashref( $ou, $query );
}

########################################
# Usage         : my $href = _get_ldap_as_hashref($ou, $query);
# Purpose       : Perform an LDAP query on the given ou.
# Returns       : A hash reference of the search results.
# Parameters    : $ou = the ou portions of the ldap base
#               : $query = the ldap query string
# Notes         : Results are cached by ou,query for repeat use.

sub _get_ldap_as_hashref {
    my $ou    = shift;
    my $query = shift;

    # passthru for cached repeat query
    if ( defined $ldap_cache->{$ou}->{$query} ) {
        return $ldap_cache->{$ou}->{$query};
    }

    my $ldap = _ldap_bind() || return;
    my $search_res = $ldap->search(
        base   => "$ou,$LDAPBASE",
        filter => $query,
    );
    if ( $search_res->code ) {
        $last_error = 'ldap search failed: ' . $search_res->error;
        return;
    }
    my $entry = $search_res->pop_entry();    # assumes only one entry
    if ( !$entry ) {
        $last_error = 'ldap search returned no matches';
        return;
    }
    my $href = {};
    foreach my $attr ( $entry->attributes ) {
        my $aref = $entry->get_value( $attr, asref => 1 );
        if ( @$aref == 1 ) {
            $href->{$attr} = $aref->[0];
        }
        else {
            $href->{$attr} = $aref;
        }
    }

    # cache the result
    $ldap_cache->{$ou}->{$query} = $href;

    return $href;
}

########################################
# Usage         : my $ldap = _ldap_bind();
# Purpose       : Connect and bind to ldap server
# Returns       : A Net::LDAP object.
# Parameters    : None.
# Notes         : Uses the cached handle if found.

sub _ldap_bind {

    # use the cached handle if found
    if ( defined $ldap_handle ) {
        return $ldap_handle;
    }

    $ldap_handle = Net::LDAP->new($LDAPSERV);
    if ( !$ldap_handle ) {
        $last_error = 'failed to connect to ldap server';
        return;
    }
    my $bind_res = $ldap_handle->bind;    # anonymous bind
    if ( !$bind_res ) {
        $last_error  = 'failed to bind to ldap';
        $ldap_handle = undef;
        return;
    }

    return $ldap_handle;
}

########################################
# Usage         : my $cn = _strip_dn( $dn );
# Purpose       : Extract the cn from an ldap DN string.
# Returns       : The extracted cn, if found.
# Parameters    : The raw ldap DN string.

sub _strip_dn {
    my $dn = shift;

    if ( $dn =~ /cn=(.*?),(?:ou|dc)=/ ) {
        return $1;
    }

    # failed return nothing
    return;
}

1;    # Magic true value required at end of module
__END__

=head1 NAME

HesiodLDAP - Perform Hesiod-like lookups using LDAP


=head1 VERSION

This document describes HesiodLDAP version 1.0.0


=head1 SYNOPSIS

    use HesiodLDAP;

    # get passwd entry for a userid
    my $passwd = resolve( $userid, 'passwd');

    # get all common entries (passwd, filsys, pobox, grplist) for a userid
    my $hes = resolve( $userid );

    # get all common entries (gid, uid) that match a given id number
    my $hes = resolve( $idnumber );


=head1 DESCRIPTION

This module was designed to be a drop-in replacement for the older
Hesiod.pm package, which was built as simply a wrapper on the hesiod
library function resolve(). This module provides the same functionality,
but uses calls to the newer LDAP database instead. We plan to end
support for hesiod on campus at some point, so this module should
provide a painless upgrade path.

To upgrade from Hesiod.pm, simply replace the use line:

    use Hesiod;

with the use line for this module:

    use HesiodLDAP;

Assuming the module has been properly installed, there should be
no further steps.

The various get_SOMETHING_hashref() functions are also available
for developing new applications. These provide access to the raw LDAP
data, which should allow the programmer to skip parsing the various
fields back out of the hesiod-formatted data returned by resolve().


=head1 VARIABLES

=head2 @HesiodLDAP::types

An array of the Hesiod types available to be queried.

This variable cannot be exported.

=head2 $Hesiod::LDAP::last_error

A string value explaining the last module failure, if any.

This variable cannot be exported.


=head1 FUNCTIONS

=head2 my $hes = resolve( $identity, $type )

Performs a Hesiod-like query on the LDAP database to resolve the 
given $type of data for the given $identity. 

$identity can be a userid, a uid number, or a hostname.

$type can be one of @HesiodLDAP::types. If $type is omitted, 
then all available types will be returned.

The result is a single string value representing the requested 
type of data. If $type is omitted, the result string may contain
multiple lines of data, each prefixed with the type. The format
should match that of the 'hes' command line program.

This function is exported by default.

=head2 my $hashref = get_accounts_hashref( $user );

Query the LDAP accounts data for the given user. $user can be a username
or a user id number. Returns a reference to a hash of the LDAP data if
a matching user is found. 

The result hash should contain these keys:

=over

=item C<< cn >> - one version of full name

=item C<< gecos >> - another version of full name

=item C<< gidNumber >> - group id

=item C<< homeDirectory >> - full afs path

=item C<< loginShell >> - default shell 

=item C<< mailHost >> - mail server host

=item C<< memberNisNetgroup >> - list of groups

=item C<< ncsuMUAProtocol >> - mail protocol 

=item C<< ncsuTwoPartName >> - simplified name

=item C<< uid >> - username

=item C<< uidNumber >> - user id number

=back

This function may be exported.

=head2 my $hashref = get_groups_hashref( $group );

Query the LDAP groups data for the given group. $group can be a groupname
or a group id number. Returns a reference to a hash of the LDAP data if
a matching group is found. 

The result hash should contain these keys:

=over

=item C<< cn >> - groupname

=item C<< description >> - description string

=item C<< gidNumber >> - group id number

=back

This function may be exported.

=head2 my $hashref = get_hosts_hashref( $hostname );

Query the LDAP host-cluster data for the given host. $hostname must be a
fully-qualified hostname for this function. Returns a reference to a
hash of the LDAP data if a matching hostname is found.

The result hash should contain these keys:

=over

=item C<< cn >> - fully-qualified hostname

=item C<< ncsuAllowNetgroup >> - list of allowed groups

=item C<< ncsuDenyNetgroup >> - list of denied groups

=back

This function may be exported.

=head2 my $hashref = get_mounts_hashref( $locker );

Query the LDAP mounts data for the given locker. $locker must be a
locker name string. Returns a reference to a hash of the LDAP data if a
matching mount is found.

The result hash should contain these keys:

=over

=item C<< cn >> - the locker name

=item C<< ncsuAFSPath >> - the full afs path to the locker

=item C<< ncsuAddable >> - a yes/no flag

=item C<< ncsuMountPath >> - the prefered mount path in /ncsu

=back

This function may be exported.

=head1 DIAGNOSTICS

=over

=item C<< identity contains invalid characters >>

Returned by resolve when the $identity that is passed contains
any characters other than (a-z, 0-9, -, _, .).

=item C<< identity was not provided >>

Returned by resolve when an empty $identity is passed as the first
parameter to the function.

=item C<< failed to resolve %s as a hostname >>

Returned by the resolver for the cluster type, when the requested
identity does not resolve in a simple DNS query. This message is an
ignorable warning when the identity being checked is not expected to be
a hostname at all.

=item C<< type '%s' is not recognized >>

Returned by resolve when a $type passed as the second paramemter 
to the function does not match one of the known types found in 
@HesiodLDAP::types.

=item C<< ldap data missing: %s: %s >>

Returned by the individual type resolvers, whenver the LDAP data that 
they get back is missing one or more fields that they need to 
construct the hesiod-like data string. The error message contains 
a short version of which LDAP table was consulted, and a list of 
which values were required for the function.

=item C<< ldap search failed: %s >>

Returned by the internal ldap search function whenever a query fails
because of an LDAP error. The LDAP error message is included in the error
string.

=item C<< ldap search returned no matches >>

Returned by the internal ldap search function whenever a query succeeds but
returns no results that matched the query.

=item C<< failed to connect to ldap server >>

Returned by the internal ldap connect function when a connection could not
be made to the ldap server. This error probably indicates a network problem
rather than a module error.

=item C<< failed to bind to ldap >>

Returned by the internal ldap connect function when a connection was made
but the anonymous bind to the service could not be completed.

=back


=head1 CONFIGURATION AND ENVIRONMENT

HesiodLDAP requires no configuration files or environment variables.


=head1 DEPENDENCIES

Net::LDAP - Lightweight Directory Access Protocol


=head1 INCOMPATIBILITIES

Note that this module uses an anonymous bind to the NCSU LDAP server.
Anonymous access to account data is only allowed from within the NCSU
IP-address space. Privacy blocks may prevent this module from working
properly from off-campus hosts.


=head1 BUGS AND LIMITATIONS

This module supports data that can be extracted from LDAP tables that
are intended to replace the Hesiod database. 

These types should be fully supported:

=over

=item C<< passwd >>

=item C<< filsys >>

=item C<< pobox >>

=item C<< gid >>

=item C<< uid >>

=item C<< grplist >>

=item C<< group >>

=back

This type is partially supported:

=over

=item C<< cluster >>

The Hesiod database included partial hostnames in the cluster data. The
LDAP data is keyed only on fully-qualified hostname. The resolver for 
cluster will attempt to expand an identity passed to it into a full 
hostname. Your results may vary, depending on the configuration of
your local DNS support.

=back

These types are not used in the old Hesiod data, so they have not
been ported to the LDAP database either. They were found in the source
code for the 'hes' binary, and so they have placeholders in this module.

=over

=item C<< sloc >>

=item C<< pcap >>

=item C<< service >>

=item C<< lpralias >>

=back

Please report any bugs or feature requests to the author.


=head1 AUTHOR

Charles J. Brabec  C<< <brabec@ncsu.edu> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2007, NC State University.  All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
